#include <stdio.h>
#include <errno.h>
#include <string>
#include <strstream>

const int numelem=256; // Anzahl Elemente in der Grundmenge
// 0 ist nullelement. 1 ist einselement. dann folgen aufsteigend die anderen bis zu numelem-1

int addtab[numelem][numelem]; // left+right
int multab[numelem][numelem]; // left*right

int subtab[numelem][numelem]; // left-right
int divtab[numelem][numelem]; // left/right

inline int lenint(int a)
{
  int res=32;
  int val=0x80000000;
  
  while ( ((a&val)==0) && (res!=0) ) {
    val>>=1;
    res--;
  };
  return res;
};

int mod256(int a)
{
  int d;
  switch(numelem) {
  case 4: d = 0x7; break;    
  case 8: d = 0xb; break;
  case 16: d = 0x1f; break;
  case 32: d = 0x25; break;
  case 256: d = 0x11b; break;
  default: printf("for numelem=%d i have no polynom in modulo calculation please adjust sourcecode\n",numelem);
    return 0; break;
  };
  
  int lend = lenint(d);
  
  for (;;) {
    int lena = lenint(a), myd;
    if (lena < lend) return a;
    if (lena > lend) myd = d << lena-lend; else myd = d;
    a ^= myd;
  };
};

int mult256(int a, int b)
{
  int res=0;
  for (int curbit=0;curbit<32;curbit++) {
    if (b&(1<<curbit)) res^=a<<curbit;
  };
  return mod256(res);
};

void loadkoerper2potx()
{
  // numelem == 2^n !!!!!!

  for (int a=0;a<numelem;++a) {
    for (int b=0;b<numelem;++b) {
      addtab[a][b]=a^b;
    };
  };
  
  for (int a=0;a<numelem;++a) {
    for (int b=0;b<numelem;++b) {
      multab[a][b]=mult256(a,b);
    };
  };
};

int calcaddinverse(int a)
{
  for (int x=0;x<numelem;++x)
    if (addtab[x][a]==0) return x;
  return 0;
};

int calcmulinverse(int a)
{
  for (int x=0;x<numelem;++x)
    if (multab[x][a]==1) return x;
  return 0;
};

void calcsubtab()
{
  for (int a=0;a<numelem;++a) {
    for (int b=0;b<numelem;++b) {
      subtab[a][b]=addtab[a][calcaddinverse(b)];
    };
  };
};

void calcdivtab()
{
  for (int a=0;a<numelem;++a) {
    for (int b=0;b<numelem;++b) {
      divtab[a][b]=multab[a][calcmulinverse(b)];
    };
  };
};

// give me array of y values.
// -1 is undefined and will be left undefined
// -2 is undefined and will be calculated by me.
void fillup (int array[numelem])
{
  // vordefinierte sttzstellen.
  int numx=0;
  int x[numelem], y[numelem]; // 0..numx-1
  
  for (int a=0;a<numelem;++a)
    if (array[a] >= 0) {
      x[numx]=a;
      y[numx]=array[a];
      numx++;
    };
  
  int nenner[numelem]; // 0..numx-1

  for (int a=0;a<numx;++a) {
    nenner[a]=1;
    for (int b=0;b<numx;++b) {
      if (a==b) continue;
      nenner[a]=multab[nenner[a]][subtab[x[a]][x[b]]];
    };
  };
  
  for (int a=0;a<numelem;++a) {
    if (array[a]==-2) {
      int zaehler=1;
      for (int b=0;b<numx;++b) {
	zaehler=multab[zaehler][subtab[a][x[b]]];
      };
      
      array[a]=0;
      for (int b=0;b<numx;++b) {
	array[a]=addtab[array[a]] [multab[divtab[divtab[zaehler][subtab[a][x[b]]]][nenner[b]]][y[b]]];
      };
    };
  };
};

void fatal(const string &x)
{
  fprintf(stderr,"\n\nFatal error:\n%s\n\n",x.c_str());
  exit(1);
};

void fatalerrno(const string &x)
{
  fprintf(stderr,"\n\nFatal error:\n%s\n\nErrno = %s\n\n",x.c_str(),strerror(errno));
  exit(1);
};

int main(int argc, char *argv[])
{
  loadkoerper2potx();
  
  calcsubtab();
  calcdivtab();
  
  FILE *files[256];
  int numfiles=argc-1;
  bool issource[256];
  
  for (int a=0;a<numfiles;++a) {
    files[a]=fopen(argv[1+a],"rb");
    if (files[a]==0) {
      if (errno==ENOENT) {
	files[a]=fopen(argv[1+a],"wb");
	if (files[a]==0) fatalerrno(string("cannot open file '")+string(argv[1+a])+string("' for writing."));
	issource[a]=false;
      } else {
	fatalerrno(string("cannot open file '")+string(argv[1+a])+string("' for reading."));
      };
    } else {
      issource[a]=true;
    };
  };

  int length=-1;
  for (int a=0;a<numfiles;++a) {
    if (issource[a]) {
      fseek(files[a],0,SEEK_END);
      if (length==-1) {
	length=ftell(files[a]);
      } else {
	if (ftell(files[a])!=length) fatal("file lengths of input files mismatch");
      };
      fseek(files[a],0,SEEK_SET);
    };
  };

  int ar[256];
  for (int a=0;a<256;++a) ar[a]=-1;
  
  for (int x=0;x<length;++x) {
    for (int a=0;a<numfiles;++a) {
      if (issource[a]) {
	unsigned char v;
	fread(&v,1,1,files[a]);
	ar[a]=v;
      } else {
	ar[a]=-2;
      }
    };

    fillup(ar); // <-- fill it up !!!

    for (int a=0;a<numfiles;++a) {
      if (!issource[a]) {
	unsigned char v=ar[a];
	fwrite(&v,1,1,files[a]);
      };
    };
  };
    
  for (int a=0;a<numfiles;++a) fclose(files[a]);

  return 0;
};
